unit Moog;

(*

  Source : https://github.com/ddiakopoulos/MoogLadders

  --

  Copyright (c) 2015, Dimitri Diakopoulos All rights reserved.

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions are met:

  * Redistributions of source code must retain the above copyright notice, this
    list of conditions and the following disclaimer.

  * Redistributions in binary form must reproduce the above copyright notice,
    this list of conditions and the following disclaimer in the documentation
    and/or other materials provided with the distribution.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

  -- --

  Huovilainen developed an improved and physically correct model of the Moog
  Ladder filter that builds upon the work done by Smith and Stilson. This model
  inserts nonlinearities inside each of the 4 one-pole sections on account of the
  smoothly saturating function of analog transistors. The base-emitter voltages of
  the transistors are considered with an experimental value of 1.22070313 which
  maintains the characteristic sound of the analog Moog. This model also permits
  self-oscillation for resonances greater than 1. The model depends on five
  hyperbolic tangent functions (tanh) for each sample, and an oversampling factor
  of two (preferably higher, if possible). Although a more faithful
  representation of the Moog ladder, these dependencies increase the processing
  time of the filter significantly. Lastly, a half-sample delay is introduced for
  phase compensation at the final stage of the filter.

  References: Huovilainen (2004), Huovilainen (2010), DAFX - Zolzer (ed) (2nd ed)
  Original implementation: Victor Lazzarini for CSound5

  Considerations for oversampling:
  http://music.columbia.edu/pipermail/music-dsp/2005-February/062778.html
  http://www.synthmaker.co.uk/dokuwiki/doku.php?id=tutorials:oversampling

  --

  2017-02-16 : Adapted for use in Wren by Blue Hell

*)


interface

uses

  System.Math, KnobsUtils, KnobsConversions;


type

  TMoogFilter = class
  private
    type TMoogState = record
      FCutoff     : TSignal;
      FResonance  : TSignal;
    end;
    PMoogState = ^ TMoogState;
  private
    FState      : array of TMoogState;
    FSampleRate : TSignal;
    FPolyphony  : Integer;
  protected
    function    GetResonance( aVoice: Integer): TSignal;                                                        virtual;
    procedure   SetResonance( aVoice: Integer; aValue: TSignal);                                                virtual;
    function    GetCutOff   ( aVoice: Integer): TSignal;                                                        virtual;
    procedure   SetCutoff   ( aVoice: Integer; aValue: TSignal);                                                virtual;
  public
    constructor Create( aPolyphony: Integer; aSampleRate: TSignal);                                             virtual;
    procedure   Process( InSample: TSignalArray; var OutSample: TSignalArray);                                  virtual;
  public
    property    Polyphony                   : Integer read FPolyphony;
    property    Resonance[ aVoice: Integer] : TSignal read GetResonance write SetResonance;
    property    CutOff   [ aVoice: Integer] : TSignal read GetCutoff    write SetCutoff;
  end;

  TMoogFilterClass = class of TMoogFilter;


  THuovilainenMoog = class( TMoogFilter)
  strict private
  const
    FThermal = 0.000025;
    FComp    = 0.5;
  private
    type THuovilainenMoogState = record
      FCutoff     : TSignal;
      FResonance  : TSignal;
      FStage0     : TSignal;
      FStage1     : TSignal;
      FStage2     : TSignal;
      FStage3     : TSignal;
      FStageTanh0 : TSignal;
      FStageTanh1 : TSignal;
      FStageTanh2 : TSignal;
      FDelay0     : TSignal;
      FDelay1     : TSignal;
      FDelay2     : TSignal;
      FDelay3     : TSignal;
      FDelay4     : TSignal;
      FDelay5     : TSignal;
      FTune       : TSignal;
      FAcr        : TSignal;
      FResQuad    : TSignal;
    end;
    PHuovilainenMoogState = ^ THuovilainenMoogState;
  private
    FState : array of THuovilainenMoogState;
  protected
    procedure   SetResonance( aVoice: Integer; aValue: TSignal);                                               override;
    procedure   SetCutoff   ( aVoice: Integer; aValue: TSignal);                                               override;
  public
    constructor Create( aPolyphony: Integer; aSampleRate: TSignal);                                            override;
    procedure   Process( InSample: TSignalArray; var OutSample: TSignalArray);                                 override;
  end;


implementation


{ ========
  TMoogFilter = class
  private
    type TMoogState = record
      FCutoff     : TSignal;
      FResonance  : TSignal;
    end;
    PMoogState = ^ TMoogState;
  private
    FState      : array of TMoogState;
    FSampleRate : TSignal;
    FPolyphony  : Integer;
  public
    property    Polyphony                   : Integer read FPolyphony;
    property    Resonance[ aVoice: Integer] : TSignal read FResonance write SetResonance;
    property    CutOff   [ aVoice: Integer] : TSignal read FCutoff    write SetCutoff;
  protected
}

    function    TMoogFilter.GetResonance( aVoice: Integer): TSignal; // virtual;
    begin
      Result := FState[ aVoice].FResonance;
    end;


    procedure   TMoogFilter.SetResonance( aVoice: Integer; aValue: TSignal); // virtual;
    begin
      FState[ aVoice].FResonance := aValue;
    end;


    function    TMoogFilter.GetCutOff( aVoice: Integer): TSignal; // virtual;
    begin
      Result := FState[ aVoice].FCutoff;
    end;


    procedure   TMoogFilter.SetCutoff( aVoice: Integer; aValue: TSignal); // virtual;
    begin
      FState[ aVoice].FCutoff := aValue;
    end;


//  public

    constructor TMoogFilter.Create( aPolyphony: Integer; aSampleRate: TSignal); // virtual;
    begin
      FPolyphony  := aPolyphony;
      FSampleRate := aSampleRate;
      SetLength( FState, Polyphony);
    end;


    procedure   TMoogFilter.Process( InSample: TSignalArray; var OutSample: TSignalArray); // virtual;
    var
      zz : Integer;
    begin
      for zz := 0 to Polyphony - 1
      do OutSample[ zz] := InSample[ zz]; // Wire bridge .. this is an abstract filter doing nothing.
    end;



{ ========
   THuovilainenMoog = class( TMoogFilter)
  private
    FStage0     : TSignal;
    FStage1     : TSignal;
    FStage2     : TSignal;
    FStage3     : TSignal;
    FStageTanh0 : TSignal;
    FStageTanh1 : TSignal;
    FStageTanh2 : TSignal;
    FDelay0     : TSignal;
    FDelay1     : TSignal;
    FDelay2     : TSignal;
    FDelay3     : TSignal;
    FDelay4     : TSignal;
    FDelay5     : TSignal;
    FTune       : TSignal;
    FAcr        : TSignal;
    FResQuad    : TSignal;
  protected
}

    procedure   THuovilainenMoog.SetResonance( aVoice: Integer; aValue: TSignal); // virtual;]begin
    begin
      FState[ aVoice].FResonance := aValue;
      FState[ aVoice].FResQuad   := FState[ aVoice].FResonance * FState[ aVoice].FAcr;
    end;


    procedure   THuovilainenMoog.SetCutoff( aVoice: Integer; aValue: TSignal); // virtual;
    var
      fc  : TSignal;
      f   : TSignal;
      fc2 : TSignal;
      fc3 : TSignal;
      fcr : TSignal;
    begin
      FState[ aVoice].FCutoff  := aValue;
      fc                       := FState[ aVoice].FCutoff / FSampleRate;
      f                        := fc * 0.5;                                   // oversampled twice
      fc2                      := fc * fc;
      fc3                      := fc2 * fc;
      fcr                      := 1.8730 * fc3 + 0.4955 * fc2 - 0.6490 * fc + 0.9988;
      FState[ aVoice].FAcr     := -3.9364 * fc2 + 1.8409 * fc + 0.9968;
      FState[ aVoice].FTune    := ( 1.0 - exp( - (( 2 * PI) * f * fcr))) / FThermal;
      FState[ aVoice].FResQuad := FState[ aVoice].FResonance * FState[ aVoice].FAcr;
    end;


//  public

    constructor THuovilainenMoog.Create( aPolyphony: Integer; aSampleRate: TSignal); // override;
    var
      zz    : Integer;
      State : PHuovilainenMoogState;
    begin
      inherited;
      SetLength( FState, Polyphony);

      for zz := 0 to Polyphony - 1
      do begin
        State := @ FState[ zz];
        State^.FCutOff    := 329.6;
        State^.FResonance := 0.10;
      end;
    end;


    procedure   THuovilainenMoog.Process( InSample: TSignalArray; var OutSample: TSignalArray); // virtual;
    var
      zz     : Integer;
      State  : PHuovilainenMoogState;
      InVal  : TSignal;
      PreVal : TSignal;
    begin
      for zz := 0 to Polyphony - 1
      do begin
        State  := @ FState[ zz];
        PreVal := Normalize( InSample[ zz]);
        InVal  := Normalize( PreVal - State^.FResQuad * ( State^.FDelay5 - FComp * InSample[ zz]));

        State^.FStage0     := Normalize( State^.FDelay0 + State^.FTune * LookupTanh( InVal * FThermal) - State^.FStageTanh0);
        State^.FDelay0     := State^.FStage0;
        State^.FStageTanh0 := Normalize( LookupTanh( State^.FStage0 * FThermal));
        State^.FStage1     := Normalize( State^.FDelay1 + State^.FTune * ( State^.FStageTanh0 - State^.FStageTanh1));
        State^.FDelay1     := State^.FStage1;
        State^.FStageTanh1 := Normalize( LookupTanh( State^.FStage1 * FThermal));
        State^.FStage2     := Normalize( State^.FDelay2 + State^.FTune * ( State^.FStageTanh1 - State^.FStageTanh2));
        State^.FDelay2     := State^.FStage2;
        State^.FStageTanh2 := Normalize( LookupTanh( State^.FStage2 * FThermal));
        State^.FStage3     := Normalize( State^.FDelay3 + State^.FTune * ( State^.FStageTanh2 - State^.FDelay3 * FThermal));
        State^.FDelay3     := State^.FStage3;
        State^.FDelay5     := 0.5 * ( State^.FStage3 + State^.FDelay4);
        State^.FDelay4     := State^.FStage3;

        InVal := Normalize( PreVal - State^.FResQuad * ( State^.FDelay5 - FComp * InSample[ zz]));

        State^.FStage0     := Normalize( State^.FDelay0 + State^.FTune * LookupTanh( InVal * FThermal) - State^.FStageTanh0);
        State^.FDelay0     := State^.FStage0;
        State^.FStageTanh0 := Normalize( LookupTanh( State^.FStage0 * FThermal));
        State^.FStage1     := Normalize( State^.FDelay1 + State^.FTune * ( State^.FStageTanh0 - State^.FStageTanh1));
        State^.FDelay1     := State^.FStage1;
        State^.FStageTanh1 := Normalize( LookupTanh( State^.FStage1 * FThermal));
        State^.FStage2     := Normalize( State^.FDelay2 + State^.FTune * ( State^.FStageTanh1 - State^.FStageTanh2));
        State^.FDelay2     := State^.FStage2;
        State^.FStageTanh2 := Normalize( LookupTanh( State^.FStage2 * FThermal));
        State^.FStage3     := Normalize( State^.FDelay3 + State^.FTune * ( State^.FStageTanh2 - State^.FDelay3 * FThermal));
        State^.FDelay3     := State^.FStage3;
        State^.FDelay5     := 0.5 * ( State^.FStage3 + State^.FDelay4);
        State^.FDelay4     := State^.FStage3;

        OutSample[ zz] := LookupTanh( State^.FDelay5);
      end;
    end;


end.

